import collections
import functools


def unique(key):
    seen_keys = set()

    @functools.wraps(unique)
    def _inner(item):
        if isinstance(item, collections.Mapping):
            key_value = item[key]
        else:
            key_value = getattr(item, key)

        if key_value in seen_keys:
            return None

        seen_keys.add(key_value)
        return item

    return _inner



import contextlib
import inspect
import traceback


class StatusTracker(object):
    def __init__(self, *names):
        self.parent = None

        self.percentages = {
            name.split("_", 1)[1]: {}
            for name in names
            if name.startswith("percentage_")
            }

        self.status_data = {
            name: 0 if name.endswith("_count") else None
            for name in names
            if name not in self.percentages
            }

        self.error_list = []
        self.task_list = []

    def get_stats(self):
        # Calculate percentages
        returner = {}
        for key, value in self.status_data.items():
            if isinstance(value, dict):
                value = {
                    k: v() if callable(v) else v
                    for k, v in value.items()
                    if v is not None
                }
            elif callable(value):
                value = value()
            elif not value:
                continue
            returner[key] = value

        for percentage, subdict in self.percentages.items():
            if not subdict:
                continue

            first, second = subdict["first"], subdict["second"]
            if callable(first):
                first = first()
            else:
                first = self.status_data[first]

            if callable(second):
                second = second()
            else:
                second = self.status_data[second]

            if 0 in (first, second):
                continue
            percent_done = int((int(first) / int(second)) * 100)
            if percent_done > 100:
                raise RuntimeError("You messed your percentage variables up: {0} & {1}".format(*data))
            returner["percentage_" + percentage] = {"percent": percent_done, "data": (first, second)}

        returner["subtasks"] = [t.get_stats() for t in self.task_list]
        returner["error_list"] = [e[0] for e in self.error_list]
        return returner

    @contextlib.contextmanager
    def subtask(self, *names):
        task = StatusTracker(*names)
        task.parent = self
        self.task_list.append(task)
        yield task
        self.task_list.remove(task)

    def func(self, key_name, func):
        self.status_data[key_name] = func

    def percentage(self, key, first, second):
        self.percentages[key]["first"] = first
        self.percentages[key]["second"] = second

    def set(self, key, value):
        self.status_data[key] = value

    def inc(self, key):
        self.status_data[key] += 1

    def error(self, ex, data=None):
        traceback.print_exc()
        self.error_list.append((ex, data))
        if self.parent:
            self.parent.error_list.append((ex, data))


class StatusMixin(object):
    status_props = set()

    def __init__(self):
        self.status = StatusTracker(*self.get_status_props())

    @classmethod
    def get_status_props(cls):
        props = set()

        for klass in inspect.getmro(cls):
            for prop in getattr(klass, "status_props", []):
                props.add(prop)

        return props



from asyncio import coroutine, get_event_loop
import io

from aiopipes import Pipeline
from aiopipes.pipeio import Output, IterableIO, FileIO
from . import TestIO
import pytest


def test_pipeline_raises_runtimetimerror(pipeline, run):
    with pytest.raises(RuntimeError):
        pipeline._convert_to_io(None, _raise=True)


def test_pipeline_concat(pipeline, test_io):
    pipe = pipeline | (lambda x: x)
    pipe < test_io

    assert isinstance(pipe, Pipeline)
    assert pipe.input is test_io
    assert pipe.output is test_io
    assert len(pipe.pipes) == 1


def test_io(pipeline):
    input, output = TestIO(), TestIO()
    pipeline < input
    pipeline > output

    assert pipeline.input is input
    assert pipeline.output is output


def test_shorthand_io(pipeline):
    input, output = range(10), io.StringIO()

    pipeline < input
    pipeline > output

    assert isinstance(pipeline.input, IterableIO)
    assert isinstance(pipeline.output, FileIO)


def test_return(pipeline, run):
    """
    Test that aiopipes can just return a single value instead of using an output
    """
    @coroutine
    def adder_return(input):
        return input + 1

    pipeline = pipeline | adder_return

    input = IterableIO(range(10))
    output = TestIO()

    pipeline < input
    pipeline > output

    run(pipeline.start())

    assert output.q == list(range(1, 11))


def test_pipes(pipeline, test_io, run):
    @coroutine
    def adder(input, output):
        yield from output(input + 1)

    new_pipeline = pipeline | adder

    input = IterableIO(range(10))

    new_pipeline < input

    run(new_pipeline.start())

    assert test_io.q == [i + 1 for i in range(10)]

    # Add another adder
    new_pipeline = pipeline | adder | adder
    new_pipeline < input
    input.reset()
    test_io.reset()

    run(new_pipeline.start())

    assert test_io.q == [i + 2 for i in range(10)]

    # Use lambda
    new_pipeline = pipeline | (lambda x: x + 1)
    new_pipeline < input
    input.reset()
    test_io.reset()

    run(new_pipeline.start())

    assert test_io.q == [i + 1 for i in range(10)]



from asyncio import coroutine, iscoroutine, wait, FIRST_COMPLETED, get_event_loop
try:
    from asyncio import ensure_future
except ImportError:
    from asyncio import async as ensure_future

from collections import Iterable
from io import TextIOBase
import inspect
import time

from aiopipes import Input, Output, IterableIO, FileIO
from .pipeio import IOFinished
from .status import StatusMixin


class Runnable(StatusMixin):
    def __init__(self, input=None, output=None):
        self.input = self._convert_to_io(input, _raise=False)
        self.output = self._convert_to_io(output)
        self.concurrency = 1
        self.started = None
        self.worker_futures = []

        super().__init__()

    @property
    def name(self):
        return self.__class__.__name__

    def log(self, msg):
        print(msg)

    @property
    def runtime(self):
        if self.started is None:
            return 0

        return time.time() - self.started

    def run_until_complete(self, ev=None):
        ev = ev or get_event_loop()

        ev.run_until_complete(self.start())

    @coroutine
    def start(self):
        @coroutine
        def runner_task():
            try:
                yield from self._run()
            except IOFinished:
                return
            except Exception as ex:
                self.status.error(ex)

        if not self.concurrency:
            # Just run a single runner_task don't bother with multiple stuff.
            fut = runner_task()
            self.worker_futures = [fut]
            try:
                yield from fut
            finally:
                if self.output:
                    yield from self.output.close()

        self.worker_futures = [ensure_future(runner_task()) for _ in range(self.concurrency)]
        self.started = time.time()

        try:
            while True:
                done, pending = yield from wait(self.worker_futures, return_when=FIRST_COMPLETED)
                self.worker_futures = list(pending)
                if not pending:
                    break
        finally:
            if self.output:
                yield from self.output.close()

    @coroutine
    def _run(self):
        raise NotImplementedError()

    def _convert_to_io(self, value, _raise=False):
        if isinstance(value, (Input, Output)):
            return value
        elif isinstance(value, TextIOBase):
            return FileIO(value)
        elif isinstance(value, Iterable):
            return IterableIO(value)
        elif _raise:
            raise RuntimeError("Cannot type {0} to input or output".format(type(value)))

    def __lt__(self, input):
        self.input = self._convert_to_io(input)
        return self

    def __gt__(self, output):
        self.output = self._convert_to_io(output)
        return self

    def parallel(self, concurrency):
        self.concurrency = concurrency
        return self


class _Continue(object):
    def __init__(self, data, done=None, max=None):
        self.data = data
        self.max = max
        self.done = done


class FunctionRunner(Runnable):
    def __init__(self, func, input=None, output=None):
        self.func = func
        super().__init__(input, output)

    @property
    def name(self):
        return self.func.__name__

    def _extract_params(self, func):
        func_args = set()

        while True:
            args = inspect.getfullargspec(func)
            for arg in args.args:
                func_args.add(arg)

            if hasattr(func, "__wrapped__"):
                func = func.__wrapped__
            else:
                break

        return func_args

    def _get_params(self, func, allowed_args=None):
        params = self._extract_params(func)
        if allowed_args is None:
            return params
        return params.intersection(allowed_args)

    def _get_param_values(self):
        @coroutine
        def _output(data):
            yield from self.output.write(data)

        return {
            "output": _output,
            "_continue": _Continue
        }

    @coroutine
    def _run(self):
        param_values = self._get_param_values()
        func_params = self._get_params(self.func, set(param_values.keys()))

        while True:
            data = yield from self.input.read()
            if data is None:
                pass

            params = {p: param_values[p] for p in func_params if p in param_values}

            with self.status.subtask("percentage_done", "done_count", "max_count") as subtask:
                subtask.percentage("done", "done_count", "max_count")

                while True:
                    try:
                        result = self.func(data, **params)
                        if iscoroutine(result):
                            result = yield from result
                        subtask.inc("done_count")

                    except Exception as ex:
                        subtask.error(ex, data)
                    else:
                        if isinstance(result, _Continue):
                            data = result.data
                            if result.max:
                                subtask.set("max_count", result.max)
                            if result.done:
                                subtask.set("done_count", result.done)
                            continue

                        if result is not None:
                            yield from self.output.write(result)

                    break



from aiopipes.runner import FunctionRunner
import functools
from asyncio import coroutine
from decorator import decorator


def test_get_args():
    def test(obj, output):
        pass

    runner = FunctionRunner(test)
    assert runner._get_params(test, {"output"}) == {"output"}


def test_lambda_args():
    runner = FunctionRunner(lambda x: x)
    assert runner._get_params(runner.func, {}) == set()


def test_get_wrapped_args():
    @decorator
    def dec_no_args(func, *args, **kwargs):
        return func(*args, **kwargs)

    @dec_no_args
    @coroutine
    def test_function(obj, output):
        return

    runner = FunctionRunner(test_function)
    assert runner._get_params(test_function, {"output"}) == {"output"}

    @decorator
    def dec_with_args(func, *args, output, **kwargs):
        return func(*args, output=output, **kwargs)

    @dec_with_args
    @coroutine
    def test_function2(output, obj):
        return

    runner = FunctionRunner(test_function2)
    assert runner._get_params(test_function2, {"output"}) == {"output"}


from aiopipes.pipeio import FileIO
from io import StringIO
import json


class UnclosableStringIO(StringIO):
    def close(self, *args, **kwargs):
        return


def test_fileio_format(pipeline, run):
    out = UnclosableStringIO()
    inp = UnclosableStringIO()

    def func(d):
        return {'number': d['number'] + 1}

    pipeline = pipeline | func
    pipeline < inp
    pipeline > out

    for i in range(10):
        inp.write(json.dumps({"number": i}) + "\n")

    inp.seek(0)

    run(pipeline.start())

    out.seek(0)

    output = []
    for line in out.readlines():
        output.append(json.loads(line))

    assert output == [{'number': i+1} for i in range(10)]


def test_fileio(tmpdir, pipeline, run):
    out_file = tmpdir.join("output").ensure()

    with out_file.open(mode="w") as fd:
        out = FileIO(fd)

        pipeline = pipeline | (lambda x: x)
        pipeline < range(10)
        pipeline > out

        run(pipeline.start())

    with out_file.open(mode="r") as fd:
        read = [int(l.strip()) for l in fd]
        assert read == list(range(10))



from asyncio import coroutine, sleep

try:
    from asyncio import ensure_future
except ImportError:
    from asyncio import async as ensure_future

from .pipeline import Pipeline
from .runner import Runnable
from .pipeio import Output, FileIO
import shutil


class BaseMonitor(Runnable):
    def __init__(self, obj: Pipeline, output: Output = None):
        if not isinstance(obj, Pipeline):
            raise RuntimeError("{name} cannot monitor {o_name}".format(
                name=self.__class__.__name__,
                o_name=obj
            ))

        self.pipe = obj
        self.monitor_future = None

        super().__init__(None, output)

    @coroutine
    def monitor(self, future):
        self.monitor_future = ensure_future(future)
        yield from self.start()


    @coroutine
    def _run(self):
        while True:
            yield from sleep(1)
            if self.monitor_future.done():
                return

            try:
                pipe_stats = [
                    {
                        "input": r.input.status.get_stats(),
                        "output": r.output.status.get_stats(),
                        "task": r.status.get_stats(),
                        "name": r.name,
                        "futures": r.worker_futures
                    }
                    for r in self.pipe.pipes]

            except Exception as e:
                self.status.error(e)
                return

            monitor_info = {
                "name": self.pipe.name,
                "runtime": self.pipe.runtime,
                "pipes": pipe_stats,
                "futures": self.pipe.worker_futures
            }
            yield from self.display(monitor_info)

    @coroutine
    def display(self, info):
        raise NotImplementedError()


class ConsoleMonitor(BaseMonitor):
    def _convert_to_io(self, value, _raise=False):
        io = super()._convert_to_io(value, _raise)
        if isinstance(io, FileIO):
            io.raw_strings = True
        return io

    @coroutine
    def display(self, info):
        columns, lines = shutil.get_terminal_size()
        shutil.get_archive_formats()

        outputs = []

        def w(msg, **extra):
            outputs.append(msg.format(i=info, **extra))

        def pre(i=0):
            if i == 0:
                return " " * 10
            return "{:-<10}".format((" " * i) + "|")

        def prog(done, current=None, max=100, waiting=False):
            small_done = (int(done * 2 / 10) or 1)
            whitespace = " " * (20 - small_done)
            end = "{current:3}/{max}".format(current=current or done, max=max) if not waiting else "  -/-"
            return "[{mark:=>{small_done}}{whitespace}] {end}".format(mark="*",
                                                                      done=done,
                                                                      end=end,
                                                                      small_done=small_done,
                                                                      whitespace=whitespace)

        w("{i[name]}: {pipe_len} pipes", pipe_len=len(info["pipes"]))
        w("Runtime: {i[runtime]:.0f} seconds")

        for sub_pipe in info["pipes"]:
            errors = "[Errors: {}]".format(len(sub_pipe["task"]["error_list"]))\
                if sub_pipe["task"].get("error_list", None) else ""
            w(pre(1) + " {p[name]} {errors}", p=sub_pipe, errors=errors)
            inp = sub_pipe["input"]
            if "percentage_done" in inp:
                percent = "{inp[percentage_done][percent]:3}% done".format(inp=inp)
                current, max = inp["percentage_done"]["data"]
            elif "read_count" in inp:
                percent, current, max = "", inp["read_count"], "?"
            else:
                percent, current, max = "", "?", "?"

            w(pre() + " Input: {current}/{max} read. {percent}", inp=inp, current=current, max=max, percent=percent)
            if "percentage_done" in inp:
                w(pre() + " " + prog(inp["percentage_done"]["percent"], current, max))

            if sub_pipe["task"]["subtasks"]:
                w(pre() + " Tasks:")

            for subtask in sub_pipe["task"]["subtasks"]:
                if "percentage_done" in subtask:
                    current, max = subtask["percentage_done"]["data"]
                    w(pre(2) + " " + prog(subtask["percentage_done"]["percent"], current=current, max=max))

        yield from self.output.write("\n".join(outputs))



from collections import Iterable
import inspect
from asyncio import coroutine, wait

try:
    from asyncio import ensure_future
except ImportError:
    from asyncio import async as ensure_future

from . import QueueIO
from aiopipes.runner import Runnable
from .runner import FunctionRunner


class Pipeline(Runnable):
    def __init__(self, name, input=None, output=None, pipes: Iterable = None):
        self._name = name
        self.pipes = pipes or []
        super().__init__(input, output)

    @property
    def name(self):
        return self._name

    def __or__(self, other):
        if not isinstance(other, Runnable):
            if inspect.isfunction(other):
                other = FunctionRunner(other)
            else:
                raise RuntimeError("Cannot convert {type} to a pipe".format(type=type(other)))

        return self.__class__(
            self.name,
            self.input,
            self.output,
            self.pipes + [other]
        )

    @coroutine
    def _run(self):
        if not self.pipes:
            self.pipes = [FunctionRunner(lambda x: x)]  # Make the pipeline a no-op

        # Hook all our aiopipes together
        internal_ios = [QueueIO() for _ in self.pipes]

        for idx, pipe in enumerate(self.pipes):
            if idx != 0:
                pipe < internal_ios[idx]

            if idx + 1 != len(internal_ios):
                pipe > internal_ios[idx + 1]

        self.pipes[0] < self.input
        self.pipes[-1] > self.output

        self.worker_futures = [
            ensure_future(pipe.start()) for pipe in self.pipes
            ]

        yield from wait(self.worker_futures)

    def __repr__(self):
        return "<Pipeline {name}>".format(name=self.name)



from setuptools import setup

setup(
    name='aiopipes',
    version='0.3',
    packages=['aiopipes'],
    url='',
    license='',
    author='Orf',
    author_email='tom@tomforb.es',
    description='Asyncio pipe structures',
    install_requires=["asyncio", "decorator"],
)



from asyncio import coroutine, Future, Queue as ioQueue


class QueueDone(Exception):
    def __init__(self, future: Future):
        self.future = future


class Queue(object):
    def __init__(self, queue: ioQueue=None):
        self._queue = queue or ioQueue()
        self.finished = False

    @property
    def items_queued(self):
        return self._queue.qsize()

    @coroutine
    def get_object(self):
        if self.finished:
            return

        obj = yield from self._queue.get()

        if isinstance(obj, QueueDone):
            for getter in self._queue._getters:
                getter.set_result(None)

            obj.future.set_result(True)
            self.finished = True
        else:
            return obj

    @coroutine
    def put_object(self, object):
        yield from self._queue.put(object)

    @coroutine
    def close(self):
        future = Future()

        yield from self.put_object(
            QueueDone(future)
        )

        return future


from asyncio import coroutine
import json
from collections import Iterable, Sized
import io
import os

from .queue import Queue
from .status import StatusMixin

__all__ = ["Input", "Output", "FileIO", "QueueIO"]


class IOFinished(Exception):
    pass


class Input(StatusMixin):
    status_props = {"read_count", "closed", "error_count"}
    @coroutine
    def read(self):
        raise NotImplementedError()


class Output(StatusMixin):
    status_props = {"write_count", "closed", "error_count"}

    @coroutine
    def write(self, data):
        raise NotImplementedError()

    @coroutine
    def close(self):
        raise NotImplementedError()


class FileIO(Input, Output):
    status_props = {"percentage_read"}

    def __init__(self, fd: io.TextIOBase, raw_strings=False):
        self.fd = fd
        self.raw_strings = raw_strings
        super().__init__()

        if isinstance(fd, io.FileIO):
            try:
                os.path.getsize(fd.name)
            except os.error:
                pass
            else:
                self.status.percentage("read", lambda: self.fd.tell(), lambda: os.path.getsize(fd.name))

    @coroutine
    def read(self):
        line = self.fd.readline()

        if not line:
            yield from self.close()
            raise IOFinished()

        try:
            return json.loads(line)
        except Exception as e:
            self.status.error(e)
            raise IOError("Could not load JSON object")
        finally:
            self.status.inc("read_count")

    @coroutine
    def write(self, data):
        try:
            if not self.raw_strings or not isinstance(data, str):
                data = json.dumps(data)
            self.fd.write(data + "\n")
            self.status.inc("write_count")
        except Exception as e:
            self.status.error(e)
            raise IOError("Could not encode JSON object")

    @coroutine
    def close(self):
        self.status.set("closed", True)
        self.fd.close()


class QueueIO(Input, Output):
    status_props = {"queued", "percentage_done"}

    def __init__(self, queue: Queue = None):
        self.queue = queue or Queue()
        super().__init__()
        self.status.func("queued", lambda: self.queue.items_queued)
        self.status.percentage("done", "read_count", "write_count")

    @coroutine
    def write(self, data):
        self.status.inc("write_count")
        yield from self.queue.put_object(data)

    @coroutine
    def read(self):
        d = yield from self.queue.get_object()
        if d is None:
            raise IOFinished()
        self.status.inc("read_count")
        return d

    @coroutine
    def close(self):
        self.status.set("closed", True)
        return (yield from self.queue.close())


class IterableIO(Input):
    status_props = {"left_count", "percentage_done"}

    def __init__(self, it: Iterable):
        self.iterable = it
        self.it = iter(self.iterable)
        super().__init__()

        if isinstance(self.iterable, Sized):
            self.status.set("left_count", len(self.iterable))
            self.status.percentage("done", "read_count", "left_count")

    @coroutine
    def read(self):
        try:
            o = next(self.it)
            self.status.inc("read_count")
        except StopIteration:
            self.status.set("closed", True)
            raise IOFinished()

        return o

    def reset(self):
        self.it = iter(self.iterable)
        self.status.set("read_count", 0)
